#!/usr/bin/env python3

# Install this with `pip3 install beautifulsoup4`
from bs4 import BeautifulSoup, NavigableString, Comment
import os, sys, re, zlib, unicodedata, sqlite3
from copy import copy
from tempfile import NamedTemporaryFile
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import urlparse, parse_qs

oed_dir = os.path.abspath(os.getenv("oed_dir"))
oed_base = os.getenv("oed_base").rstrip("/")
assert oed_base.startswith("/") or not oed_base

headings_shift = int(os.getenv("headings_start")) - 1
assert 0 <= headings_shift <= 3

entry_headings_shift = int(os.getenv("entry_headings_start")) - 1
assert 0 <= entry_headings_shift <= 3

testing = bool(os.getenv("testing"))

schema = """
pragma foreign_keys = on;
create table dictionary(
	entry_id integer primary key,
	search_key text not null,
	headword text not null,
	contents text not null
);
create index dictionary_index on dictionary(search_key);
create table lemmas(
	lemma text not null,
	entry_id integer not null,
	foreign key(entry_id) references dictionary(entry_id)
);
create table extra_pages(
	path text primary key,
	html text not null
);
"""

this_dir = os.path.abspath(os.path.dirname(__file__))

conn = None

def read_oed_file(path, mode="r"):
	if not os.path.isabs(path):
		path = os.path.join(oed_dir, path)
	if os.path.isdir(path):
		files = os.listdir(path)
		files.sort()
		return [os.path.join(path, f) for f in files]
	with open(path, mode) as f:
		return f.read()

def raw_file_path(name):
	return os.path.join(this_dir, "raw", name)

def temp_file(mode="w+b"):
	return NamedTemporaryFile(mode, delete=False, dir=this_dir, suffix=".tmp")

# entities.txt maps entity names to sequences of Unicode characters. Characters
# are given by names, separated with commas, when there is no character
# corresponding to a given entity, a placeholder name is given instead,
# preceded by the character `*`.
def load_entities():
	path = os.path.join(this_dir, "entities.txt")
	seen = set()
	entities = []
	f = open(path)
	for line in f:
		name, value = line.strip().split(None, 1)
		assert name not in seen
		seen.add(name)
		if value.startswith("*"):
			value = value.lstrip("*")
			if os.path.exists(os.path.join(this_dir, "svg", name + ".svg")):
				value = f'<img src="{oed_base}/svg/{name}.svg" class="oed-svg" alt="{value}"/>'
			else:
				value = f'<span class="oed-unsupported">{value}</span>'
		else:
			buf = []
			for c in value.split(","):
				c = unicodedata.lookup(c.strip())
				buf.append(c)
			value = "".join(buf)
		entities.append(("&%s;" % name, value))
	f.close()
	return dict(entities)

def extract_entries():
	data = read_oed_file("oed.t", "rb")
	f = temp_file("w")
	while data:
		d = zlib.decompressobj(-15)
		try:
			t = d.decompress(data).decode("ascii")
		except (zlib.error, UnicodeDecodeError):
			t = ""
		if len(t) < 2 or not t.startswith('#') or not t.endswith('#'):
			data = data[1:]
			continue
		for entry in t.strip("#").split("#"):
			f.write("%s\n" % entry)
		data = d.unused_data
	f.flush()
	os.rename(f.name, raw_file_path("entries.txt"))
	f.close()

def extract_headwords():
	data = read_oed_file("hw.t", "rb")
	data = zlib.decompress(data).decode()
	data = data.replace("^", "\n")
	f = temp_file("w")
	f.write(data)
	f.flush()
	os.rename(f.name, raw_file_path("headwords.txt"))
	f.close()

def extract_search_keys():
	data = read_oed_file("ky.t", "rb")
	data = zlib.decompress(data).decode()
	data = data.strip("#").replace("#", "\n") + "\n"
	f = temp_file("w")
	f.write(data)
	f.flush()
	os.rename(f.name, raw_file_path("search_keys.txt"))
	f.close()

def emend(data):
	# <nu> and <dn>: numerator and denominator appear within a single
	# entry, an editing error; note that <dn> are not opened/closed
	# correctly
	data = data.replace("1,  1 / 2</dn>, <nu>1</nu><dn>4</dn>, <nu>1</nu><dn>8</dn>, <nu>1</nu><dn>16",
		"1, &half;, &oneon4;, &oneon8;, &oneon16;")
	# single occurrence of xcn; this is an error. the number was probably
	# an internal reference, but the reference is incorrect (no "Babism" in
	# the dictionary).
	data = data.replace("<xr><xcn>00016130</xcn><x>Babism</x></xr>", "Babism")
	return data

def preformat_tables(record):
	# for formatting tables
	# <l>: used in only two entries, within tables, to mark text that should
	# be centered.
	record = record.replace('<l just="centre"><i>Old English.</i><br><sc><tab>Singular.<tab>Dual.<tab>Plural.</sc>',
		'<br><sc><tab>Singular.<tab>Dual.<tab>Plural.</sc><l just="centre"><i>Old English.</i>')
	record = record.replace('<br><sc><tab>Singular.<tab>Dual.<tab>Plural.</sc>',
		'<tr><th></th><th>Singular.</th><th>Dual.</th><th>Plural.</th></tr>')
	tmp = '<tr><th>Sing.</th><th>Masc.</th><th>Fem.</th><th>Neut.</th><th>Plural</th></tr>'
	record = record.replace('<br><sc>Sing.<tab>Masc.<tab>Fem.<tab>Neut.<tab>Plural</sc>', tmp)
	record = record.replace('<br><sc>sing.<tab>masc.<tab>fem.<tab>neut.<tab>plural</sc>', tmp)
	record = record.replace('<br><sc>Sing.<tab>Masc.<tab>Fem.<tab>Neut.<tab>Plural.</sc>', tmp)
	for word in "Old", "Middle", "Modern":
		record = record.replace('<l just="centre"><i>%s English.</i>' % word,
			'<tr class=oed-table-subheading><td></td><td colspan="3" style="text-align:center">' + \
			'<span style="text-align:center;font-style:italic">' + \
			'%s English.</span></td></tr>' % word)
	# need to do that for tables, otherwise beautifulsoup closes <tab> weirdly
	# and this screws up parsing of the table's contents
	record = record.replace("<tab>", "<tab/>")
	return record

# <textformat tabstops="100,200,300,400,500,600,700,800">...</textformat>
# encapsulates a table. each row begins with a <br>, columns are
# separated with <tab>.
def format_tables(soup):
	for tf in soup.find_all("textformat"):
		table = soup.new_tag("table")
		table["class"] = "oed-table"
		child = next(tf.children)
		first_val = True
		while child is not None:
			if child.name == "br":
				row = soup.new_tag("tr")
				table.append(row)
				column = soup.new_tag("td")
				row.append(column)
			elif child.name == "tab":
				column = soup.new_tag("td")
				row.append(column)
			elif child.name in ("tr", "th"):
				table.append(copy(child))
			else:
				if first_val:
					first_val = False
					# in the following case the table in
					# fact represents an enumeration
					if isinstance(child, NavigableString) and str(child)[0].isdigit():
						table["class"] = "oed-enum"
				column.append(copy(child))
			child = child.next_sibling
		tf.replace_with(table)
		table.insert_before(soup.new_tag("br"))

def format_quotes(soup):
	# <w>: = work. title of a work from which a quotation is extracted.
	# in italics.
	for w in soup.find_all("w"):
		w.name = "i"
	# <qp>: quotations paragraphs? encapsulates a set of quotations. there
	# are three non-breaking spaces at the beginning of each quotation
	# (&nb;&nb;&nb;), should kill them and put one quotation per line.
	for qp in soup.find_all("qp"):
		qp.name = "div"
		qp["class"] = "oed-quotations"
	# <pqp>: a paragraph within a <pq>. some definitions are followed by
	# several paragraphs of quotations. always preceded by <br><br>, thus
	# this should not be interpreted as <p>.
	for pqp in soup.find_all("pqp"):
		pqp.unwrap()
	# <q>: encapsulates a quote with author, date, work, text, in purple
	# also have <q ed="yes">, don't see the diff.; quotes added later on?
	for q in soup.find_all("q"):
		q.attrs.clear()
		q.name="p"
	# <d>: date of a quotation (free form). we have <d sb="yes"> if the
	# quote (not just the date) is surrounded with brackets; there can be
	# several quotes within a single brackets pair. can ignore the "sb"
	# attribute.
	for d in soup.find_all("d"):
		d.name = "b"
	# <s>: author of a quotation, in small caps. sometimes not an author
	# name, though.
	for s in soup.find_all("s"):
		s.name = "span"
		s["class"] = "oed-small-caps"
	# <qt>: = quotation text. no visual effect.
	for qt in soup.find_all("qt"):
		qt.unwrap()

def format_entities(record, entities=None):
	record = record.replace("&nb;&nb;&nb;", "")
	if not entities:
		entities = load_entities()
	return re.sub(r"&[^;]+;", lambda x: entities[x.group(0)], record)

def format(record, entities=None):
	lemmas = set()
	# there isn't necessarily a single root node (sometimes there is an
	# entry, and additions follow outside of the <e> element, so we need
	# one.
	record = '<div>%s</div>' % record
	# <hl>: used just before an addition/draft. an horizontal line. always of
	# the form <hl>______________________________</hl>
	record = record.replace("<hl>______________________________</hl>", "<hr>")
	record = preformat_tables(record)
	record = format_entities(record, entities)
	record = record.replace('<br><br>', '')

	soup = BeautifulSoup(record, "html.parser")

	# <e>: encapsulates a main entry.
	# some entries have <e type="nq">, they tend to be much smaller. not sure
	# what this attribute means.
	# <sube>: encapsulates a sub entry (for additions and draft). occurs
	# after <e> is closed. there can be several <sube> at the end of a <e>.
	# if an addition, we have:
	#   <e><upd>Additions 1993</upd><br><br><kg><hg text="add93" vol="1">...</hg>...
	#   <e><upd>Additions 1993</upd><br><br><kg><hg text="add93" vol="2">...</hg>...
	#   <e><upd>Additions 1997</upd><br><br><kg><hg text="add97" vol="3">...</hg>...
	# also have:
	#   <e><upd>Draft entry June 2003</upd><br><br><kg><hg text="new">
	# otherwise just:
	#   <e><kg><hg text="oed">...</hg>...
	# <upd> can appear further in an entry, see e.g. hoop and hoot, v.
	for upd in soup.find_all("upd"):
		upd.name = "h%d" % (2 + entry_headings_shift)
	for e in soup.find_all("e"):
		e.unwrap()
	for sube in soup.find_all("sube"):
		sube.unwrap()
	# <kg>: = key group? corresponds to the headword + whether it is a
	# noun, verb., etc. and sometimes extra info. in fact there can be
	# several headwords per entry, probably the reason they're within a
	# <kg>. within a <kg> we have:
	# <hw>: headword (possibly with extra signs like a cross),
	# <ps>: person? we have n., adj., stuff like that. in italics.
	# <tf>: additional observations on the headword, no special effect
	# at the end of a kg, apparently always have two <br><br>
	for kg in soup.find_all("kg"):
		kg.name = "h%d" % (1 + entry_headings_shift)
		kg["class"] = "oed-header"
	for hg in soup.find_all("hg"):
		hg.unwrap()
	for hw in soup.find_all("hw"):
		hw.name = "b"
	for ps in soup.find_all("ps"):
		ps.name = "i"
	for tf in soup.find_all("tf"):
		tf.unwrap()
	# <hm>: = headword mark? marks a superscript number like in
	# -ine<hm>5</hm>
	# <x>kist</x> <ps>n.<hm>1</hm></ps>
	for hm in soup.find_all("hm"):
		hm.name = "sup"
	# <lg>: lines below kg and before definitions. lg because comes after kg?
	for lg in soup.find_all("lg"):
		lg.name = "div"
		lg["class"] = "oed-entry-extra"
	# <dg>: word formation. no visual effect.
	for dg in soup.find_all("dg"):
		dg.unwrap()
	# <sg>: entry body, with several definitions and quotes
	for sg in soup.find_all("sg"):
		sg.name = "div"
		sg["class"] = "oed-entry-body"
	# <pg>: phonetics group? followed by an opening bracket, then
	# one or more <ph> = phonetics for each possible pronunciation,
	# then closing bracket
	for pg in soup.find_all("pg"):
		pg.name = "div"
		pg["class"] = "oed-phonetics"
	# <gk>: greek text
	for gk in soup.find_all("gk"):
		gk.unwrap()
	# <def>: a definition. preceded by 2 <br>, unless we have
	# <def br="no">, which occurs when the definition occurs on the
	# same line as the previous one.
	def_class = "oed-def-block"
	for def_ in soup.find_all("def"):
		if "br" in def_.attrs and def_.previous_sibling == " ":
			# try to merge with the previous def
			left = def_.previous_sibling.previous_sibling
			if left.name == "div" and left["class"] == def_class:
				left.append(def_.previous_sibling)
				left.append(def_)
				def_.unwrap()
				continue
		def_.name = "div"
		def_.attrs.clear()
		def_["class"] = def_class
	# tables
	format_tables(soup)
	# <num>: paragraph num. <pnum>: paragraph num. within a <def>, only the
	# last part followed by a dot e.g. for II.2 we have <pnum>2.</pnum>
	# <pnum> is immediately followed by <anum>, which gives us
	# the full definition number without a trailing dot e.g.
	# <anum>II.2</anum> <anum> is not displayed within the app., but
	# appears when doing a copy/paste from within the app., probably a bug.
	for anum in soup.find_all("anum"):
		anum.decompose()
	for pnum in soup.find_all("pnum"):
		pnum.unwrap()
	for num in soup.find_all("num"):
		num.name = "b"
	# <spg>: = spellings group. mentions other spellings, free format.
	# no visual effect.
	for spg in soup.find_all("spg"):
		spg.unwrap()
	# <et>: no idea what this is, no visual effect
	for et in soup.find_all("et"):
		et.unwrap()
	# <b>: used around each word that is an alternative spelling of the
	# entry word. e.g. for betwixt we have <b>-twux</b>, <b>betweohs</b>
	# etc. represented in bold face. the tagging is far from accurate. <b>
	# is employed where <bf> should be.
	# <su>: superscript
	for su in soup.find_all("su"):
		su.name = "sup"
	# <i>: italics
	# <sc>: small capitals
	for sc in soup.find_all("sc"):
		sc.name = "span"
		sc["class"] = "oed-small-caps"
	# <ss>: sans-serif
	for ss in soup.find_all("ss"):
		ss.name = "span"
		ss["class"] = "oed-sans-serif"
	# <susc>: like superscript but a bit lower (the small text doesn't
	# reach above a capital in the normal font. also in small caps.
	for susc in soup.find_all("susc"):
		susc.name = "span"
		susc["class"] = "oed-suscrit"
	# <il>: marks lemmas within an entry. formatted italics + bold face.
	# with an <il>, there is sometimes an <shw> = shared word? that isn't
	# displayed in an interface e.g.:
	# <il>go<shw>&sp;aback</shw></il>, <il>come a-back</il>
	# becomes "go, come a-back".
	# we must extract lemmas like that somewhere else.
	for il in soup.find_all("il"):
		lemma = copy(il)
		for shw in lemma.find_all("shw"):
			shw.unwrap()
		for r in lemma.find_all("r"):
			r.name = "span"
			r["class"] = "oed-joining-word"
		lemma = str(lemma)
		lemma = lemma.split('>', 1)[1].rsplit('<', 1)[0]
		lemmas.add(lemma)
		il.name = "span"
		il["class"] = "oed-lemma"
	for shw in soup.find_all("shw"):
		shw.decompose()
	# <r>: marks things like 'and', 'et al.', 'or', 'prop.'. which should
	# be displayed in a normal font while their enclosing element is bold
	# or italic. must be done after extracting lemmas.
	for r in soup.find_all("r"):
		r.name = "span"
		r["class"] = "oed-joining-word"
	# <n>: a note within a definition; smaller font size; introduced like
	# that: <br>&nb;&nb;&nb;<n>...</n>
	for n in soup.find_all("n"):
		n.name = "div"
		n["class"] = "oed-note"
	# <bl>: secondary headwords. bl = bold large? represented in bold face.
	for bl in soup.find_all("bl"):
		bl.name = "b"
	format_quotes(soup)
	# <a>: all of them are of the form <a href='event:x$X'> where $X is the
	# id of an entry, which is its index in the dictionary, counting from 0.
	for a in soup.find_all("a"):
		id = int("".join(c for c in a["href"] if c.isdigit()))
		# +1 so that we can map that to a row id in sqlite
		a["href"] = "%s/ids/%d" % (oed_base, id + 1)
	# <xr>: cross-reference
	# like <xr><x>nine</x> <ps>adj.</ps></xr>
	# within that, <x> is the bare word, put in small caps.
	for xr in soup.find_all("xr"):
		xr.unwrap()
	for x in soup.find_all("x"):
		x.name = "span"
		x["class"] = "oed-small-caps"
	# <in>: marks a subscript number; within chem. formulas
	# <insc>: I cannot verify what this is. probably in+sc,
	# so subscript + small caps
	for in_ in soup.find_all("in"):
		in_.name = "sub"
	for insc in soup.find_all("insc"):
		insc.name = "sub"
		insc["class"] = "oed-small-caps"
	# <ov>: within math and chem. formulas, no visible effect
	for ov in soup.find_all("ov"):
		ov.unwrap()
	# <cs>: for math and chemistry. cs = change script?
	# seems to split the line into two rows.
	# within that we can have <co> which is what to display above
	# also <cu>, which is what do display below.
	# also <bco>,<bcu>: likewise, apparently for groups
	# also <bsco>,<bscu>?
	# also <bo>: bracketer overline?
	# also <bu>: bracketed underline like |______| below the line
	# can't render these, so at least make the structure apparent
	for name in ("cs", "co", "cu", "bco", "bcu", "bsco", "bscu", "bo", "bu"):
		for tag in soup.find_all(name):
			span = soup.new_tag("span")
			span["class"] ="oed-unsupported"
			span.string = "<%s>" % name
			tag.insert_before(span)
			span = soup.new_tag("span")
			span["class"] ="oed-unsupported"
			span.string = "</%s>" % name
			tag.insert_after(span)
			tag.unwrap()
	# should be more accurate about this, but this potentially messes up
	# only ~1500 entries
	for br in soup.find_all("br"):
		br.decompose()
	return str(soup.div), lemmas

def load_entries():
	with open(raw_file_path("entries.txt")) as f:
		data = f.read()
	return emend(data).splitlines()

def format_entries():
	entities = load_entities()
	lines = load_entries()
	data = None
	entries = []
	for i, entry in enumerate(lines, 1):
		print("[%06d/%d]" % (i, len(lines)), file=sys.stderr)
		entry, lemmas = format(entry, entities)
		yield entry, lemmas

def shift_headings(soup):
	for level in (3, 2, 1):
		for tag in soup.find_all("h%d" % level):
			tag.name = "h%d" % (level + headings_shift)

def parse_biblio_file(soup):
	entries = []
	for entry in soup.find_all("div", {"class": "bibentry"}):
		anchors = []
		links = []
		prev = entry.previous_sibling
		if prev.name == "a":
			assert "name" in prev.attrs, prev
			anchors.append(prev["name"].strip())
		entry["class"] = "oed-bib-entry"
		for a in entry.find_all("a"):
			if "name" in a.attrs:
				anchors.append(a["name"])
			else:
				assert "href" in a.attrs
				letter, anchor = re.match(r"^(?:(?:/help/bib/)?oed2-([a-z])\d?\.html)?(#.+)$", a["href"]).groups()
				if letter:
					anchor = "%s%s" % (letter, anchor)
				a["href"] = anchor
				links.append(anchor)
		for div in entry.find_all("div", {"class": "detail"}):
			div["class"] = "oed-bib-detail"
		# there can be several headers per bibentry. see
		# under Davenant, Sir William.
		for div in entry.find_all("div", {"class": "header"}):
			div["class"] = "oed-bib-header"
		entries.append(str(entry))
	return entries

def parse_biblio():
	entities = load_entities()
	entries = []
	prev_letter = ""
	for path in read_oed_file("Help/bib"):
		name = os.path.basename(path)
		if not name.startswith("oed2-"):
			continue
		letter = name[len("oed2-")]
		if letter != prev_letter and entries:
			conn.execute("insert into extra_pages values(?, ?)",
				("%s/bib/%s" % (oed_base, prev_letter), "".join(entries)))
			entries.clear()
		prev_letter = letter
		text = read_oed_file(path)
		text = text.replace('<img align="absbottom" alt="{Epsilon}" border="0" src="/graphics/parser/gifs/mb/Epsilon.gif"/><img align="absbottom" alt="{asper}" border="0" src="/graphics/parser/gifs/mb/asper.gif"/>',
			'<img alt="{Easper}" src="foobar"/>')
		text = text.replace('<img align="absbottom" alt="{Eta}" border="0" src="/graphics/parser/gifs/mb/Eta.gif"/><img align="absbottom" alt="{asper}" border="0" src="/graphics/parser/gifs/mb/asper.gif"/>',
			'<img alt="{Hasper}" src="foobar"/>')
		for k, v in entities.items():
			text = text.replace(k, v)
		soup = BeautifulSoup(text, "html.parser")
		for span in soup.find_all("span", {"class": "last"}):
			span["class"] = "oed-small-caps"
		for img in soup.find_all("img"):
			alt = img["alt"].strip("{}")
			repl = entities["&%s;" % alt]
			img.replace_with(repl)
		entries.extend(parse_biblio_file(soup))
	conn.execute("insert into extra_pages values(?, ?)",
		("%s/bib/%s" % (oed_base, prev_letter), "".join(entries)))
	return entries

def parse_biblio_index():
	text = read_oed_file("Help/bib/index.html")
	text = text.split('Return to main Help</a></a></p>')[1]
	text = text.split('<p><a href="../mainhelp')[0]
	text = text.replace("B<small>ACON</small>", '<span class="oed-small-caps">Bacon</span>')
	soup = BeautifulSoup(text, "html.parser")
	soup.find(class_="toc").decompose()
	t = soup.find("table")
	t.attrs.clear()
	t.next_sibling.next_sibling.decompose()
	div = soup.new_tag("div")
	div["class"] = "oed-directory"
	soup.h1.insert_after(div)
	for letter in "abcdefghijklmnopqrstuvwxyz":
		a = soup.new_tag("a")
		a["href"] = "bib/%s" % letter
		a.string = letter.upper()
		div.append(a)
		if letter != "z":
			a.insert_after(" ")
	shift_headings(soup)
	conn.execute("insert into extra_pages values(?, ?)",
		("%s/bib" % oed_base, str(soup)))

def parse_symbols():
	entities = load_entities()
	text = read_oed_file("Help/symbols.html")
	text = text.split("Return to main Help</a></p>")[1]
	text = text.split('<div class="non-pdf">')[0]
	tmp = 'alt="{em}" border="0" width="14" height="15"> '
	text = text.replace(tmp,  tmp.strip() + "\N{NARROW NO-BREAK SPACE}")
	text = text.replace('<b>:</b><img src=', '<b>:\N{NARROW NO-BREAK SPACE}</b><img src=')
	soup = BeautifulSoup(text, "html.parser")
	for img in soup.find_all("img"):
		img.replace_with(entities["&%s;" % img["alt"][1:-1]])
	for name in ("table", "tr"):
		for tag in soup.find_all(name):
			tag.attrs.clear()
	for span in soup.find_all("span"):
		span.unwrap()
	soup.find("a").unwrap()
	shift_headings(soup)
	conn.execute("insert into extra_pages values(?, ?)",
		("%s/symbols" % oed_base, str(soup).replace("— ", "—")))

pronunciation_titles = {
	"FOREIGN AND NON-SOUTHERN": 'Foreign and Non-Southern',
	"SHORT": "Short",
	"NASAL": "Nasal",
	"LONG": "Long",
	"DIPHTHONGS, etc.": "Diphtongs, etc.",
	'</span></span><img': '<img', # spurious closing tags
}
def parse_pronunciation():
	entities = load_entities()
	text = read_oed_file("Help/pronunciation-2e.html")
	text = text.split('<p><a href="mainhelp.html">Return to main Help</a></p>')[1]
	text = text.split('<div class="infobox">')[0]
	for k, v in pronunciation_titles.items():
		text = text.replace(k, v)
	soup = BeautifulSoup(text, "html.parser")
	for img in soup.find_all("img"):
		img.replace_with(entities["&%s;" % img["alt"][1:-1]])
	for name in ("table", "tr", "td", "th"):
		for tag in soup.find_all(name):
			tag.attrs.clear()
	for span in soup.find_all("span"):
		span.unwrap()
	for font in soup.find_all("font"):
		font.name = "span"
		font.attrs.clear()
		font["class"] = "oed-small-caps"
	for table in soup.find_all("table"):
		for tr in table.find_all("tr"):
			tds = tr.find_all("td") or tr.find_all("th")
			if len(tds) > 1:
				tds[1].decompose()
				tds[3].decompose()
			else:
				td = tds[0]
				td["colspan"] = "2"
	shift_headings(soup)
	conn.execute("insert into extra_pages values(?, ?)",
		("%s/ipa" % oed_base, str(soup)))

def parse_abbreviations():
	soup = BeautifulSoup(read_oed_file("Help/abbrevs/combined.html"), "html.parser")
	table = soup.find_all("table")[1]
	pairs = []
	for row in table.find_all("tr"):
		pair = row.find_all("td")
		if len(pair) < 2:
			continue
		assert len(pair) == 2
		pairs.append((pair[0].get_text(), pair[1].get_text()))
	buf = ["<h%d>Abbreviations</h%d>" % (headings_shift + 1, headings_shift + 1)]
	buf.append("<table>")
	for p in pairs:
		buf.append("<tr><td>*%s</td><td>%s</td></tr>" % p)
	buf.append("</table>")
	conn.execute("insert into extra_pages values(?, ?)", (f"{oed_base}/abbrs", "".join(buf)))

def load_headwords():
	with open(raw_file_path("headwords.txt")) as f:
		headwords = f.read()
	headwords = re.sub(r"\{([1-9])\}", r"<sup>\1</sup>", headwords)
	headwords = headwords.replace("&", "&amp;")
	headwords = format_entities(headwords, None)
	return headwords.splitlines()

def make_dictionary():
	headwords = load_headwords()
	with open(raw_file_path("search_keys.txt")) as f:
		search_keys = f.read()
	search_keys = search_keys.replace(
		"898number", "898").replace(
		"360degree", "360").replace(
		"900number", "900")
	search_keys = search_keys.splitlines()
	entries = testing and load_entries() or format_entries()
	itor = enumerate(zip(search_keys, headwords, entries), 1)
	for i, (search_key, headword, entry) in itor:
		if not testing:
			entry, lemmas = entry
		conn.execute("""insert into dictionary(entry_id, search_key, headword, contents)
			values(?, ?, ?, ?)""", (i, search_key, headword, entry))
		if not testing:
			for l in lemmas:
				conn.execute("insert into lemmas values(?, ?)", (l, i))

def make_database():
	funcs = [
		parse_biblio_index,
		parse_biblio,
		parse_symbols,
		parse_pronunciation,
		parse_abbreviations,
	]
	f = temp_file()
	global conn
	conn = sqlite3.connect(f.name)
	conn.execute("pragma journal_mode=OFF")
	conn.executescript(schema)
	for func in funcs:
		func()
	make_dictionary()
	conn.commit()
	conn.close()
	os.rename(f.name, os.path.join(this_dir, "oed.sqlite"))
	f.close()


max_entries_per_page = 15

class Handler(BaseHTTPRequestHandler):

	def fail(self):
		self.send_response(404, "Not Found")
		self.end_headers()
		self.wfile.write(b"not found")

	def print(self, s):
		self.wfile.write(s.encode())

	def print_header(self):
		path = os.path.join(this_dir, "oed.css")
		with open(path) as f:
			css = f.read()
		self.print(f'''<!DOCTYPE html><html><head><meta charset="utf-8">
			<style>{css}</style></head><body>
			<h1>Oxford English Dictionary</h1>
			''')

	def print_search_form(self):
		base = oed_base or "/"
		self.print(f'''<form action="{base}" method="get">
			<input type="text" name="q" autofocus>
			<input type="submit" value="Search">
			</form>''')

	def print_trailer(self):
		root = oed_base or "/"
		self.print(f'''
			<footer>
			<menu>
			<li><a href="{root}">Home</a></li>
			<li><a href="{oed_base}/symbols">Symbols and conventions</a></li>
			<li><a href="{oed_base}/abbrs">Abbreviations</a></li>
			<li><a href="{oed_base}/bib">Bibliography</a></li>
			</menu>
			</footer>
			</body></html>''')

	def print_entries(self, entries, offset):
		if not entries:
			self.send_response(200, "OK")
			self.end_headers()
			self.print_header()
			self.print_search_form()
			self.print("<p>Not found.</p>")
			self.print_trailer()
			return
		if len(entries) == 1:
			id = entries[0][0]
			self.send_response(301, "Moved Permanently")
			self.send_header("Location", f'{oed_base}/ids/{id}')
			return self.end_headers()
		self.send_response(200, "OK")
		self.end_headers()
		self.print_header()
		self.print_search_form()
		self.print('<nav><ul>')
		for id, headword, _ in entries[:max_entries_per_page]:
			self.print(f'<li><a href="{oed_base}/ids/{id}">{headword}</a></li>')
		self.print('</ul></nav>')
		if offset > 0 or len(entries) > max_entries_per_page:
			base = oed_base or "/"
			self.print('<nav class="pagination">')
			if offset > 0:
				prev = offset - max_entries_per_page
				self.print(f'<a href="{base}?o={prev}">← Previous</a> | ')
			else:
				self.print("← Previous | ")
			if len(entries) > max_entries_per_page:
				next = offset + max_entries_per_page
				self.print(f'<a href="{base}?o={next}">Next →</a>')
			else:
				self.print("Next →")
			self.print('</nav>')
		self.print_trailer()

	def print_entry(self, entry):
		if not entry:
			return self.fail()
		entry = entry[0]
		self.send_response(200, "OK")
		self.end_headers()
		self.print_header()
		self.print_search_form()
		self.print(f'<div class="oed-entry">')
		if testing:
			entry, _ = format(entry)
		self.print(entry)
		self.print('<div>')
		self.print_trailer()

	def print_search_page(self):
		self.send_response(200, "OK")
		self.end_headers()
		self.print_header()
		self.print_search_form()
		self.print_trailer()

	def do_GET(self):
		url = urlparse(self.path)
		page = conn.execute("select html from extra_pages where path=?", (url.path,)).fetchone()
		if page:
			self.send_response(200, "OK")
			self.end_headers()
			self.print_header()
			self.print(page[0])
			return self.print_trailer()
		if os.path.dirname(url.path) == f"{oed_base}/svg":
			path = os.path.join(this_dir, "svg", os.path.basename(url.path))
			with open(path, "rb") as f:
				data = f.read()
			self.send_response(200, "OK")
			self.send_header("Content-Type", "image/svg+xml")
			self.send_header("Content-Length", len(data))
			self.end_headers()
			return self.wfile.write(data)
		if url.path == (oed_base or "/"):
			if not url.query:
				return self.print_search_page()
			parts = parse_qs(url.query)
			q = parts.get("q", ["*"])[0]
			offset = int(parts.get("o", ["0"])[0])
			entries = conn.execute("""
			select rowid, headword, contents from dictionary
			where search_key glob ?
			limit (? + 1) offset ?
			""", (q, max_entries_per_page, offset)).fetchall()
			return self.print_entries(entries, offset)
		if re.match("^%s/ids/[0-9]+$" % oed_base, url.path):
			id = int(os.path.basename(url.path))
			entry = conn.execute("select contents from dictionary where rowid = ?", (id,)).fetchone()
			return self.print_entry(entry)
		return self.fail()

def run(host="localhost", port=8888):
	global conn
	conn = sqlite3.connect(os.path.join(this_dir, "oed.sqlite"))
	server = HTTPServer((host, port), Handler)
	print("running test server at %s:%s" % (host, port))
	server.serve_forever()

commands = {
	"extract-entries": extract_entries,
	"extract-headwords": extract_headwords,
	"extract-search-keys": extract_search_keys,
	"build": make_database,
	"run": run,
}

if __name__ == "__main__":
	commands[sys.argv[1]]()
